"""
Qfits.py: a module to read header information from a FITS file.

This module offers a function to read all headers in a given
FITS file and return them as a list of strings.
"""

Qfits_version = '5.4.0'

# $Id: Qfits.py,v 1.27 2005/07/21 12:07:48 yjung Exp $
# $Author: yjung $
# $Date: 2005/07/21 12:07:48 $
# $Revision: 1.27 $

import qfits

#
# Check the underlying C libray version
#
if qfits.version() != Qfits_version:
    print "warning: qfits.so and Qfits.py differ in version"
    print "         qfits.so is", qfits.version()
    print "         Qfits.py is", Qfits_version

def get_headers(filename):
    """
    This function is the main operator in this module.
    Given a file name, it returns all header informations
    in a list of lists. The top list contains one element
    per found header, i.e. if the given file has 8 extensions,
    the top list contains 9 items (1 for the main header,
    8 for the extension headers). Each list contains
    in turn all informations about the corresponding
    headers as a list of tuples like

    (key, value, comment)

    Example:

    The given FITS file contains 1 main header and 2
    extension headers given as follows:

    --- main header
    SIMPLE  = T / FITS format
    BITPIX  = 8 / Bits per pixel
    NAXIS   = 0 / No data
    ...
    END

    --- extension 1
    XTENSION = 'IMAGE' / Image extension
    BITPIX   = 16      / Bits per pixel
    ...
    END

    --- extension 2
    XTENSION = 'IMAGE' / Image extension
    BITPIX   = -32     / Bits per pixel
    ...
    END

    The returned list contains three items.
    The first one is a list corresponding to the
    main header, then come the two extension
    headers. The first list looks like:

    [ ('SIMPLE','T','FITS format'),
      ('BITPIX', '8', 'Bits per pixel'),
      ('NAXIS', '0', 'No data'),
      ...
      ('END', '', '') ]

    The extension lists look like this:

    [ ('XTENSION', 'IMAGE', 'Image extension'),
      ('BITPIX', '16', 'Bits per pixel'),
      ...
      ('END', '', '') ]

    [ ('XTENSION', 'IMAGE', 'Image extension'),
      ('BITPIX', '-32', 'Bits per pixel'),
      ...
      ('END', '', '') ]

    In summary, if you want to get all FITS header informations
    from a FITS file, call this function with:

    hdr = qfits.get_headers(filename)

    This function will raise an IOError exception if the provided
    filename does not correspond to a valid file, and None if
    the file is not FITS.

    If the load succeeded, you can access individual keywords
    with the following syntax:

    hdr[extension][row][index]

    Where:

    extension runs from 0 to N_EXT (inclusive)
    row is the row number of the card
    index is 0 for the keyword, 1 for the value, 2 for the comment.

    The number of FITS extensions found in the file is simply:
    len(hdr)-1

    """
    return qfits.get_headers(filename)

def expand_keyword(key):
    """
    This function is useful to expand a given keyword into
    its FITS equivalent. What it does is simply bring the
    word to uppercase, then expand shortFITS words to HIERARCH
    ESO notation, like:

    simple    ->  SIMPLE
    NaXiS     ->  NAXIS
    det.dit   ->  HIERARCH ESO DET DIT
    """
    return qfits.expand_keyword(key)

def datamd5(filename):
    """
    This function computes the MD5 signature of all data parts in a FITS
    file and returns a 32-char string containing the signature as an
    hexadecimal number (128 bits). It raises an IOError exception if the
    given filename does not correspond to a valid FITS file.
    """
    return qfits.datamd5(filename)

def version():
    """
    This function returns the version number of the
    underlying qfits module.
    """
    return qfits.version()

class fitsinfo:
    """
    The fitsinfo class encapsulates queries into a FITS header.

    Initialize an instance by giving the name of a file, the following
    fields are then filled in:

        - filename     - Name of the file
        - n_ext        - Number of FITS extensions in file
        - hdrs         - List of headers
        - xtnum        - Current extension number for get()

    The following methods are offered:

    * get()
      This method retrieves keyword information in a given instance.
      Provide one or more keywords to look for and this function will
      browse the hdrs list for matching keywords. This method returns a
      string if you requested only 1 keyword or a list of strings if you
      requested several keywords. A failed match returns None.

      This method will only search for keywords in the xtnum extension.

      get() supports the shortFITS notation, in which A.B.C is transformed
      into HIERARCH ESO A B C before lookup in the FITS header. Notice that
      the given strings are also converted to uppercase before the search
      is launched. See the function expand_keyword() in this module.

      Returned string values are always pretty-formatted, i.e. something
      stored into a FITS string as 'o''hara' will be returned as o'hara.

    Examples:

    q = fitsinfo('vltframe.fits')
    q.get('simple')             # returns 'T'
    q.get('simple', 'naxis')    # returns ('T', '2')

    To search for keywords in the 3rd extension:

    q = fitsinfo('wfiframe.fits')
    q.xtnum = 3
    q.get('xtension')                # returns 'IMAGE'
    q.get('det.chip3.id', 'bitpix')
    # returns ('ccd52', '16')
    
    * datamd5()
      This method computes the MD5 signature of all the data parts of the
      given FITS file and returns it as a 32-char string containing a
      hexadecimal number on 128 bits.

    """
    def __init__(self, filename):
        """
        Provide a filename to initialize the object.
        """
        self.filename = filename
        self.hdrs     = qfits.get_headers(filename)
        if not self.hdrs:
            raise IOError, "not a FITS file"
        self.n_ext    = len(self.hdrs)-1
        self.xtnum    = 0

    def get(self, *keywords):
        """
        Get a list of keyword values and return it as a list. If only one
        keyword is requested, the returned value is a single string (or
        None if the keyword was not found). If several keywords are passed,
        the returned value is a list of found values.

        The search is only performed in the current extension (xtnum field
        in the class).
        """
        retval = []
        for keyword in keywords:
            keyword = qfits.expand_keyword(keyword)
            if len(keyword)<1:
                retval.append(None)
                continue
            found=0
            for k in self.hdrs[self.xtnum]:
                if k[0] == keyword:
                    found=1
                    retval.append(k[1])
                    break
            if not found:
                retval.append(None)
        if len(retval)==1:
            return retval[0]
        return retval

    def __repr__(self):
        return "file: %s - %d extensions" % (self.filename, self.n_ext)

    def datamd5(self):
        return qfits.datamd5(self.filename)

if __name__ == '__main__':

    # This short code demonstrates the module capabilities
    
    import string, sys

    if len(sys.argv)<2:
        print "provide a FITS file name as first argument"
        raise SystemExit

    filename = sys.argv[1]
    try:
        # Try building an instance of fitsinfo
        q = fitsinfo(filename)
    except IOError:
        print "cannot read file", filename
        raise SystemExit

    # Demonstrate string representation of the instance
    print `q`
    # Compute the DATAMD5 value for this file
    print "computing DATAMD5 for", q.filename, "..."
    md5hash = q.datamd5()
    print "DATAMD5 = '"+md5hash+"'"


    # Dump the first 3 cards in each header to console to show they have
    # been correctly loaded.
    xtnum=0
    for h in q.hdrs:
        print "-"*65
        print "header beginning for %s (ext %d):" % (q.filename, xtnum)
        print "-"*65
        for card in [h[0], h[1], h[2]]:
            print "key[%s] val[%s] com[%s]" % (card[0], card[1], card[2])
        print "..."
        print "-"*65
        xtnum = xtnum+1

    #
    # And now, for something completely different...
    #
    # An interesting example: transform a list of fetched keywords into a
    # dictionary for easier parsing. The idea is to build a dictionary-like
    # object that will allow queries into FITS headers like:
    # lx, ly = d["naxis1"], d["naxis2"]
    # It also simplifies notation for searches, like:
    #
    # (Python 2.2)
    # if 'det.dit' in d:
    #     print 'DIT value is', d['det.dit']
    #
    # (earlier Python versions)
    # if 'det.dit' in d.keys():
    #     print 'DIT value is', d['det.dit']
    #
    # An alternative using exceptions:
    # try:
    #     print 'DIT value is', d['det.dit']
    # except KeyError, key:
    #     print 'DIT value cannot be found'

    # Until Python 2.1, dictionaries (as all built-in types) cannot be
    # directly subclassed but through the UserDict module.
    # Starting with Python 2.2, dictionaries can directly be subclassed
    # from the 'dict' built-in class.
    # The following code makes the appropriate import if needed.

    try:
        dict
    except:
        from UserDict import UserDict as dict

    class FITSdic(dict):
        # A query will pass the key through expand_keyword to make sure
        # we can request something using lowercase and shortFITS
        # notation.
        def __getitem__(self, key):
            return dict.__getitem__(self, qfits.expand_keyword(key))

    # Create one instance
    d = FITSdic()
    # Fill it up with values from the previously loaded main header
    for card in q.hdrs[0]:
        d[card[0]] = card[1]

    # Now show keywords stored in a dictionary.
    # Protect against keys that are not in dictionary (raises KeyError):
    print "Main header info:"
    keylist = ["simple", "bitpix", "naxis", "ra", "dec", "det.dit"]
    for k in keylist:
        try:
            print "%-12s= %s" % (k, d[k])
        except KeyError, key:
            print "%-12s= <NotFound>" % (k,)

    # That's all folks
    print "end of Qfits demo"

